/**
  *  \file client/keymaphandler.cc
  *  \brief Widget for "UseKeymap" command
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/keymaphandler.h"
#include "cpluslib/format.h"
#include "gfx/gfxcontext.h"
#include "gfx/palette.h"
#include "int/simple.h"
#include "ui/skin.h"
#include "ui/ui.h"
#include "ui/uiglobal.h"
#include "util/atom.h"
#include "util/nls.h"

namespace {
    /** Keymap handler widget.
        Waits for a keystroke.
        If it gets one, converts it into a command according to the keymap.
        If the user doesn't immediately press a key, displays a reminder. */
    class WKeymapHandler : public UIFixedWidget {
     public:
        WKeymapHandler(IntKeymapRef k);

        virtual void drawContent(GfxCanvas& can);
        virtual bool handleEvent(const UIEvent& event, bool second_pass);

        uint32_t getCommand();

     private:
        void show();
        string_t getText();

        int time;
        IntKeymapRef keymap;
        uint32_t command;
    };

    /* Pop up the keymap handler after two seconds. */
    const int POPUP_TICKS = 2 * TIMER_HZ;

    /* Font to use */
    const int FONT = FONT_NORMAL_BOLD;
}

WKeymapHandler::WKeymapHandler(IntKeymapRef k)
    : UIFixedWidget(),
      time(0),
      keymap(k),
      command(0)
{ }

void
WKeymapHandler::drawContent(GfxCanvas& can)
{
    if (time >= POPUP_TICKS) {
        const GfxRect& r = getExtent();
        GfxContext ctx;
        ctx.useCanvas(can);
        ctx.useColorScheme(GfxStandardColorScheme::instance);

        drawSolidBar(ctx, r, COLOR_TOOLTIP);

        ctx.setColor(COLOR_TOOLTIP_SHADE);
        drawHLine(ctx, r.x, r.getBottomY()-1, r.getRightX()-1);
        drawVLine(ctx, r.getRightX()-1, r.y, r.getBottomY()-2);

        ctx.setColor(COLOR_TOOLTIP_LIGHT);
        drawHLine(ctx, r.x+1, r.y, r.getRightX()-1);
        drawVLine(ctx, r.x, r.y, r.getBottomY()-2);

        ctx.setColor(COLOR_TOOLTIP_TEXT);
        ctx.useFont(*fonts[FONT]);
        outText(ctx, r.x + 3, r.y + 1, getText());

        can.update(getExtent());
    }
}

bool
WKeymapHandler::handleEvent(const UIEvent& event, bool second_pass)
{
    switch (event.type) {
     case ev_Tick:
        if (++time == POPUP_TICKS) {
            show();
        }
        return UIFixedWidget::handleEvent(event, second_pass);

     case ev_Key:
        command = (keymap->lookupCommand(event.key.code));
        if (command != 0) {
            // This key was bound. Use it.
            stop(0);
        } else if (event.key.code == SDLK_ESCAPE) {
            // Escape
            stop(0);
        } else if (event.key.code == vk_Quit) {
            // Quit
            putKey(vk_Quit);
            stop(0);
        } else if (getClassOfKey(event.key.code & ~ss_AllMod) == KeyNormal) {
            // Not bound, but something that looks like a key (i.e. not a modifier).
            if (time < POPUP_TICKS) {
                time = POPUP_TICKS;
                show();
            }
        } else {
            // Something else, e.g. Shift.
        }
        return true;

     case ev_Mouse:
        if (event.mouse.b != 0) {
            ungetEvent(event);
            stop(0);
        }
        return true;

     default:
        return UIFixedWidget::handleEvent(event, second_pass);
    }
}

inline uint32_t
WKeymapHandler::getCommand()
{
    return command;
}

void
WKeymapHandler::show()
{
    const string_t text = getText();
    GfxRect r(0, 0,
              fonts[FONT]->getTextWidth(text) + 6,
              fonts[FONT]->getTextHeight(text) + 2);
    r.centerWithin(ui_root->getExtent());
    setExtent(r);
}

string_t
WKeymapHandler::getText()
{
    return format(_("Keymap %s"), keymap->getName());
}


/******************************* Main Entry ******************************/

/** Secondary keymap (USEKEYMAP command).
    Await the secondary keystroke, and executes the command.
    \param k Keymap to use
    \param prefix Prefix argument */
void
doUseKeymap(IntKeymapRef k, int prefix)
{
    if (!k) {
        return;
    }

    // User interaction
    WKeymapHandler handler(k);
    handler.setExtent(GfxRect(0, 0, 0, 0));
    ui_root->runModal(handler);

    // Command execution
    uint32_t cmd = handler.getCommand();
    if (isAtom(cmd)) {          // implies cmd != 0
        executeStatement(atomStr(cmd), prefix);
    }
}
