/**
  *  \file client/widgets/searchobj.cc
  *  \brief Search Query Object Selector
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/searchobj.h"
#include "util/nls.h"
#include "gfx/palette.h"
#include "client/search.h"
#include "util/unicodechars.h"

/** Constructor.
    \param id Widget Id */
WSearchObjectSelector::WSearchObjectSelector(int id, int from, int to)
    : UIListbox(id, to - from, font_ems[FONT_NORMAL]*10, to - from),
      objs((1 << to) - (1 << from)),
      fromType(from),
      toType(to)
{
    sig_dblclick.add(this, &WSearchObjectSelector::onToggle);
}

/** Destructor. */
WSearchObjectSelector::~WSearchObjectSelector()
{ }

/** Set object selection (bitmask).
    \param objs Selected objects, see TUserSearchQuery */
void
WSearchObjectSelector::setObjects(uint16_t objs)
{
    objs &= (1 << toType) - (1 << fromType);
    if (this->objs != objs) {
        this->objs = objs;
        drawWidget();
    }
}

/** Get object selection (bitmask). */
uint16_t
WSearchObjectSelector::getObjects() const
{
    return this->objs;
}

void
WSearchObjectSelector::drawPart(GfxCanvas& can, int from, int to)
{
    // xref TUserSearchQuery::SearchObject
    static const char* const names[] = {
        N_("Starships"),
        N_("Planets"),
        N_("Starbases"),
        N_("Ufos"),
        N_("Others"),
        N_("Only my units"),
    };

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);

    const int mid = 2*font_ems[FONT_NORMAL];
    for (int i = from; i < to; ++i) {
        /* Prepare */
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);

        /* Draw */
        int line = getTop() + i + fromType;
        if (line >= 0 && line < int(countof(names))) {
            outTextF(ctx, area.splitX(mid), (objs & (1 << line)) ? " " UTF_CHECK_MARK : " " UTF_MIDDLE_DOT);
            outTextF(ctx, area,             _(names[line]));
        }
    }
}

bool
WSearchObjectSelector::handleEvent(const UIEvent& event, bool second_pass)
{
    if (hasState(st_Focused) && event.type == ev_Key && event.key.code == ' ') {
        sig_dblclick.raise(getId());
        return true;
    }
    return UIListbox::handleEvent(event, second_pass);
}

void
WSearchObjectSelector::onItemClick(int item, int x)
{
    if (x < 2*font_ems[FONT_NORMAL]) {
        this->objs ^= 1 << (fromType + item);
        drawWidget();
    }
}

void
WSearchObjectSelector::onToggle()
{
    this->objs ^= 1 << (fromType + getCurrent());
    drawWidget();
}
