/**
  *  \file tools/c2dump.cc
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <vector>
#include <cstdlib>
#include "tools/c2dump.h"
#include "arch/platform.h"
#include "cpluslib/format.h"
#include "cpluslib/types.h"
#include "io/dirs.h"
#include "io/io.h"
#include "util/console.h"
#include "util/except.h"
#include "util/modinit.h"
#include "util/nls.h"
#include "util/recode.h"
#include "version.h"

static const char* arg_progname;

static void
die(const string_t s)
{
    console.write(LOG_ERROR, format("%s: %s", arg_progname, s));
    std::exit(2);
}

void
DumpOutputReceiver::addField(const char* name, string_t value)
{
    addField(string_t(name), value);
}

/***************************** DumpBlockInput ****************************/

DumpBlockInput::DumpBlockInput(const char* p, std::size_t size)
    : p(p), size(size)
{ }

DumpBlockInput::DumpBlockInput(DumpBlockInput& other, std::size_t rsize)
    : p(other.p), size(0)
{
    if (rsize > other.size)
        rsize = other.size;
    size = rsize;
    other.size -= rsize;
    other.p += rsize;
}

DumpBlockInput::~DumpBlockInput()
{ }

string_t
DumpBlockInput::readByte()
{
    if (getRemainingSize() >= 1) {
        uint8 value = *p++;
        size--;
        return format("%d", value);
    } else {
        return "<missing>";
    }
}

string_t
DumpBlockInput::readWord()
{
    if (getRemainingSize() >= 2) {
        int16 value = getInt16(p);
        p += 2, size -= 2;
        return format("%d", value);
    } else {
        return "<missing>";
    }
}

string_t
DumpBlockInput::readLong()
{
    if (getRemainingSize() >= 4) {
        int32 value = getInt32(p);
        p += 4, size -= 4;
        return format("%d", value);
    } else {
        return "<missing>";
    }
}

string_t
DumpBlockInput::readCoordinate()
{
    if (getRemainingSize() >= 4) {
        string_t x = readWord();
        string_t y = readWord();
        return "(" + x + "," + y + ")";
    } else {
        return "<missing>";
    }
}

string_t
DumpBlockInput::readString(std::size_t length)
{
    if (getRemainingSize() >= length) {
        string_t s(p, length);
        p += length;
        size -= length;
        string_t::size_type p = s.find(char('\0'));
        if (p == s.npos)
            p = s.size();
        while (p > 0 && s[p-1] == ' ')
            --p;
        s.erase(p);
        return "'" + convertGameToUtf8(s) + "'";
    } else {
        return "<missing>";
    }
}

string_t
DumpBlockInput::readPascalString()
{
    uint8_t length;
    if (read(&length, 1) == 1) {
        return readString(length);
    } else {
        return "<missing>";
    }
}

std::size_t
DumpBlockInput::getRemainingSize() const
{
    return size;
}

string_t
DumpBlockInput::readUnparsed(std::size_t limit)
{
    std::size_t n = getRemainingSize();
    if (n > limit)
        n = limit;

    string_t result;
    for (std::size_t i = 0; i < n; ++i) {
        if (i != 0)
            result += ' ';
        result += format("%02X", uint8(*p));
        ++p, --size;
    }
    return result;
}

int
DumpBlockInput::peek(void* buffer, size_t size)
{
    using std::memcpy;
    if (size > this->size)
        size = this->size;
    memcpy(buffer, p, size);
    return size;
}

int
DumpBlockInput::read(void* buffer, size_t size)
{
    int got = peek(buffer, size);
    return skip(got);
}

int
DumpBlockInput::skip(size_t size)
{
    if (size > this->size)
        size = this->size;
    this->size -= size;
    this->p += size;
    return size;
}

void
dumpRemainder(DumpBlockInput& bi, DumpOutputReceiver& r)
{
    while (bi.getRemainingSize() > 0) {
        r.addUnparsedData(bi.readUnparsed(16));
    }
}

/****************************** Main Control *****************************/

namespace {
    class DumpSimpleOutputReceiver : public DumpOutputReceiver {
        int level;
     public:
        DumpSimpleOutputReceiver()
            : level(0)
            { }
        void startRecord(string_t header);
        void addField(string_t name, string_t value);
        void addUnparsedData(string_t formatted);
        void endRecord();
    };

    class DumpTypeDetect {
        std::vector<const char*> possible_types;
        DumpParser found_parser;
        string_t type_override;
        string_t file_basename;
     public:
        DumpTypeDetect();
        void setRequiredType(const char* name);
        void setFileBaseName(const char* base);
        void start();
        void match(const char* type, const char* base, DumpParser parser);

        int getNumPossibleTypes() const
            { return possible_types.size(); }
        DumpParser getParser() const
            { return found_parser; }
        const char* getType() const
            { return possible_types[0]; }
    };
}

void
DumpSimpleOutputReceiver::startRecord(string_t header)
{
    console << string_t(2*level, ' ') << header << ":\n";
    ++level;
}

void
DumpSimpleOutputReceiver::addField(string_t name, string_t value)
{
    while (name.size() < 30)
        name += ' ';
    console << string_t(2*level, ' ') << name << " = " << value << "\n";
}

void
DumpSimpleOutputReceiver::addUnparsedData(string_t formatted)
{
    console << string_t(2*level, ' ') << "Unparsed = " << formatted << "\n";
}

void
DumpSimpleOutputReceiver::endRecord()
{
    console << "\n";
    --level;
}

DumpTypeDetect::DumpTypeDetect()
    : found_parser(0), type_override(), file_basename()
{ }

void
DumpTypeDetect::setRequiredType(const char* name)
{
    type_override = strLCase(name);
}

void
DumpTypeDetect::setFileBaseName(const char* base)
{
    file_basename = strLCase(base);
}

void
DumpTypeDetect::start()
{
    possible_types.clear();
    found_parser = 0;
}

void
DumpTypeDetect::match(const char* type, const char* base, DumpParser parser)
{
    using namespace std;

    // We match this entry if type is same as specified using -t, or -t gave "auto" (or blank)
    if (type_override.size() != 0 && type_override != "auto") {
        if (type_override != type)
            return;
    } else if (base == 0) {
        // No type specified, and no basename, so we cannot accept this
        return;
    } else {
        // Type is "auto", so check basename. If this type has no basename filter,
        // reject it.
        size_t size = strlen(base);
        if (size <= file_basename.size() && file_basename.compare(0, size, base, size) != 0)
            return;
        if (file_basename.size() != size
            && file_basename[size] != '.'
            && (file_basename[size] < '0' || file_basename[size] > '9'))
            return;
    }

    // Okay, matches.
    if (parser != found_parser) {
        found_parser = parser;
        possible_types.push_back(type);
    }
}

int main(int /*argc*/, char** argv)
{
    using namespace std;
    DumpTypeDetect tydet;
    const char* last_t = 0;
    enum { NoPreference, PlayerFiles, HostFiles } opt_prefs = NoPreference;
    int retval = 0;
    int nfiles = 0;
    performInitialisation(OnStartup, mod_base_console);
    argv = Platform::convertArgumentsToUtf8(argv);
    arg_progname = *argv;
    while (const char* p = *++argv) {
        while (p[0] == '-' && p[1] == '-')
            ++p;
        if (strncmp(p, "-t", 2) == 0) {
            if (p[2] != 0) {
                last_t = &p[2];
            } else {
                if (const char* q = *++argv) {
                    last_t = q;
                } else {
                    die(_("option '-t' needs an argument."));
                }
            }
            tydet.setRequiredType(last_t);
        } else if (strcmp(p, "-p") == 0) {
            opt_prefs = PlayerFiles;
        } else if (strcmp(p, "-m") == 0) {
            opt_prefs = HostFiles;
        } else if (strcmp(p, "-help") == 0 || strcmp(p, "-h") == 0) {
            console << format(_("PCC2 File Dump Utility %s - (c) 2010-2021 Stefan Reuther\n\n"
                                "Usage:\n"
                                "  %s [-tTYPE] FILE [[-tTYPE] FILE...]\n\n"
                                "Options:\n"
                                "  -tTYPE    Select type to use to interpret following files\n"
                                "  -p        Autodetect player files\n"
                                "  -m        Autodetect host files\n"
                                "  -CCHARSET Set character set\n"
                                "\n"
                                "Report bugs to <Streu@gmx.de>\n"),
                                PCC2_VERSION,
                                arg_progname);
            console.flush();
            return 0;
        } else if (strncmp(p, "-C", 2) == 0) {
            const char* cs = p+2;
            if (!*p) {
                cs = *++argv;
                if (!cs)
                    die(_("option `-C' needs an argument (the character set)"));
            }
            CharacterSet cset = CharacterSet::getCharsetByName(cs);
            if (!cset.isValid()) {
                die(_("the specified character set is not known"));
            }
            setGameCharacterSet(cset);
        } else if (*p == '-') {
            die(format(_("invalid option specified. Use `%s -h' for help"), arg_progname));
        } else {
            // It's a file argument
            ++nfiles;
            tydet.setFileBaseName(Directory::getFileName(p).c_str());
            tydet.start();
            tydet.match("bdata", "bdata", parseBDataFile);
            tydet.match("beamspec", "beamspec", parseBeamSpec);
            tydet.match("engspec", "engspec", parseEngSpec);
            tydet.match("gendat", opt_prefs == PlayerFiles ? "gen" : 0, parseGenFile);
            tydet.match("hullspec", "hullspec", parseHullSpec);
            tydet.match("names", "planet", parseNameList);
            tydet.match("names", "storm", parseNameList);
            tydet.match("pdata", "pdata", parsePDataFile);
            tydet.match("racenm", "race", parseRaceNames);
            tydet.match("ship",  "ship", parseShipFile);
            tydet.match("target", "target", parseTargetFile);
            tydet.match("torpspec", "torpspec", parseTorpSpec);
            tydet.match("truehull", "truehull", parseTrueHull);
            tydet.match("xyplan", "xyplan", parseXYPlan);
            tydet.match("team", "team", parseTeams);
            tydet.match("vcr", "vcr", parseVcrFile);
            if (tydet.getNumPossibleTypes() == 0) {
                if (last_t) {
                    // FIXME: this also appears with "-tauto"
                    console.write(LOG_ERROR, format(_("%s: type '%s' is unknown"), p, last_t));
                } else {
                    console.write(LOG_ERROR, format(_("%s: unable to auto-detect type"), p));
                }
                retval = 1;
            } else if (tydet.getNumPossibleTypes() > 1) {
                console.write(LOG_ERROR, format(_("%s: file matches multiple types, use '-tTYPE' to specify correct one"), p));
                retval = 1;
            } else {
                try {
                    Ptr<Stream> s = Stream::openFile(p, Stream::C_READ);
                    console << "=== Dump of file '" << p << "' using type '" << tydet.getType() << "' ===\n\n";
                    DumpSimpleOutputReceiver out;
                    tydet.getParser()(*s, out);
                }
                catch (FileProblemException& e) {
                    console.write(LOG_ERROR, format("%s: %s", p, e.what()));
                }
                catch (std::exception& e) {
                    console.write(LOG_ERROR, format("%s: %s", p, e.what()));
                }
            }
        }
    }

    console.flush();
    if (!nfiles)
        die(_("No file names specified. Use 'c2dump -help' for help"));

    return retval;
}
