/**
  *  \file ui/widgets/filechoose.cc
  *  \brief File Chooser Widget
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "ui/widgets/filechoose.h"
#include "gfx/gfxcontext.h"
#include "util/nls.h"
#include "ui/msgbox.h"
#include "cpluslib/format.h"
#include "resmgr/resmgr.h"


/** Get level of a directory. FIXME: this routine is also in
    dirchoose.cc. Put it in dirs.cc? */
static int
getLevel(Ptr<Directory> dir)
{
    int level = 0;
    while (dir != 0 && level < 1000) {
        ++level;
        dir = dir->getParent();
    }
    return level;
}



/** Constructor.
    \param id Widget Id
    \param initial_directory Initial directory to be viewed
    \param wildcard Wildcard to use
    \param columns Preferred number of columns
    \param lines Preferred number of lines */
UIFileChooser::UIFileChooser(int id, Ptr<Directory> initial_directory, string_t wildcard, int columns, int lines)
    : UIScrollable(id),
      columns(columns),
      lines(lines),
      current_columns(0),
      current_column_width(0),
      icons(requestPixmap("files")),
      current_directory(initial_directory),
      file_patterns(1, wildcard)
{
    scanDirectory(initial_directory, false);
}

UIFileChooser::~UIFileChooser()
{ }

void
UIFileChooser::addPattern(const FileNamePattern& pat)
{
    file_patterns.push_back(pat);
}

void
UIFileChooser::getLayoutInfo(LayoutInfo& info)
{
    info.pref_size = info.min_size = font_sizes[FONT_NORMAL].scaledBy(columns*15, lines);
    info.flags = li_Flexible;
}

bool
UIFileChooser::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key && hasState(st_Focused) && !hasState(st_Disabled)) {
        switch (event.key.code) {
         case SDLK_UP:
         case vk_WheelUp:
            setSelectedWidget();
            scrollTo(getCurrent()-1, -1);
            return true;
         case SDLK_DOWN:
         case vk_WheelDown:
            setSelectedWidget();
            scrollTo(getCurrent()+1, +1);
            return true;
         case SDLK_HOME:
            setSelectedWidget();
            scrollTo(0, 1);
            return true;
         case SDLK_END:
            setSelectedWidget();
            scrollTo(getSize()-1, -1);
            return true;
         case SDLK_PAGEUP:
            setSelectedWidget();
            scrollTo(getCurrent() - getPageSize(), -1);
            return true;
         case SDLK_PAGEDOWN:
            setSelectedWidget();
            scrollTo(getCurrent() + getPageSize(), -1);
            return true;
         case SDLK_RIGHT:
            setSelectedWidget();
            scrollTo(getCurrent() + getExtent().h / font_heights[FONT_NORMAL], +1);
            return true;
         case SDLK_LEFT:
            setSelectedWidget();
            scrollTo(getCurrent() - getExtent().h / font_heights[FONT_NORMAL], -1);
            return true;
         case SDLK_BACKSPACE:
            setSelectedWidget();
            if (current_directory) {
                setDirectory(current_directory->getParent());
            }
            return true;
        }
    } else if (event.type == ev_Mouse && !hasState(st_Disabled) && event.mouse.b != 0) {
        if (isInside(GfxPoint(event.mouse.x, event.mouse.y))) {
            /* Clicked inside widget: figure out new position */
            setSelectedWidget();
            setFocusedWidget();

            const int dx      = current_column_width;
            const int dy      = font_heights[FONT_NORMAL];
            const int lines   = getExtent().h / dy;
            const int columns = current_columns;

            const int line   = (event.mouse.y - getExtent().y) / dy;
            const int column = (event.mouse.x - getExtent().x) / dx;
            if (line < lines && column < columns) {
                scrollTo(getTop() + line + column*lines, +1);
            }
            if (event.mouse.b & mb_Double) {
                handleEnter();
            }
            return true;
        }
        if (hasState(st_Focused) && second_pass) {
            /* We're focused, and nobody is interested in this mouse
               click: check whether it's just above our first
               displayed item, or just below our last displayed item,
               to allow mouse scrolling. */
            if (event.mouse.x >= extent.x && event.mouse.x <= extent.x + current_column_width
                && event.mouse.y < extent.y && event.mouse.y >= extent.y - font_heights[FONT_NORMAL])
            {
                setSelectedWidget();
                scrollTo(getTop() - 1, -1);
                return true;
            }
            if (event.mouse.x >= extent.getRightX() - current_column_width
                && event.mouse.x < extent.getRightX()
                && event.mouse.y >= extent.getBottomY()
                && event.mouse.y < extent.getBottomY() + font_heights[FONT_NORMAL])
            {
                setSelectedWidget();
                scrollTo(getTop() + getPageSize(), +1);
                return true;
            }
        }
    }
    return UIScrollable::handleEvent(event, second_pass);
}

void
UIFileChooser::drawContent(GfxCanvas& can)
{
    /* Compute size. Note that this may draw a partial column, but not a partial line. */
    const int dx      = current_column_width;
    const int dy      = font_heights[FONT_NORMAL];
    const int lines   = getExtent().h / dy;
    const int columns = current_columns;

    /* Graphics context */
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());

    /* Draw it */
    int item = getTop();
    for (int column = 0; column < columns; ++column) {
        int relx  = dx * column;
        int width = dx;
        if (width + relx > getExtent().w)
            width = getExtent().w - relx;
        int x = relx + getExtent().x;
        int y = getExtent().y;
        for (int line = 0; line < lines; ++line) {
            GfxRect area(x, y, width, dy);
            if (item >= getSize()) {
                /* This item does not exist */
                drawBackground(ctx, area);
            } else {
                /* Item exists */
                if (item == getCurrent()) {
                    if (getFocusState() != NoFocus) {
                        drawSolidBar(ctx, area, UIColor::tc_Static);
                        ctx.setColor(UIColor::tc_InvStatic);
                    } else {
                        drawBackground(ctx, area);
                        ctx.setColor(UIColor::tc_Static);
                        drawRectangle(ctx, area);
                    }
                } else {
                    drawBackground(ctx, area);
                    ctx.setColor(UIColor::tc_Static);
                }
                if (item < int(directories.size())) {
                    ctx.useFont(*fonts[FONT_NORMAL_BOLD]);
                    if (icons) {
                        icons->blit(ctx.getCanvas(), area.x, area.y, GfxRect(0, directories[item].isUp ? 32 : 16, 16, 16));
                    }
                    area.consumeX(20);
                    outTextF(ctx, area, directories[item].title);
                } else {
                    ctx.useFont(*fonts[FONT_NORMAL]);
                    if (icons) {
                        icons->blit(ctx.getCanvas(), area.x, area.y, GfxRect(0, 0, 16, 16));
                    }
                    area.consumeX(20);
                    outTextF(ctx, area, files[item-directories.size()]->getBaseName());
                }
                ++item;
            }
            y += dy;
        }
    }

    /* Done */
    can.update(getExtent());
}

void
UIFileChooser::scrollTo(int pos, int /*corr*/)
{
    /* don't do anything if we're empty */
    if (getSize() == 0)
        return;

    /* use adjustPosition to find out new position */
    int oldPos = getCurrent();
    int oldTop = getTop();
    setCurrent(pos);
    adjustPosition();
    if (getCurrent() != oldPos || getTop() != oldTop) {
        sig_moved.raise(getId());
        drawWidget();
    }
}

void
UIFileChooser::onResize(GfxRect oldsize, bool by_layout)
{
    updateSize();
    UIScrollable::onResize(oldsize, by_layout);
}

void
UIFileChooser::onStateChange(int state, bool enable)
{
    if (state & st_Focused)
        drawWidget();
    UIScrollable::onStateChange(state, enable);
}

/** Set displayed directory.
    \param directory New directory. Can be null to display the list of roots. */
void
UIFileChooser::setDirectory(Ptr<Directory> directory)
{
    Ptr<Directory> previous = current_directory;
    current_directory = directory;
    scanDirectory(directory, true);
    showDirectory(previous);
    sig_directory_changed.raise();
}

/** Handle user input.
    - when given a pattern, changes the wildcard used for filtering
    - when given a directory, changes to that directory
    - when given a file name (may or may not exist), makes that the result
    - combinations are allowed

    \retval true input processed, caller should clear input line
    \retval false input could not be processed */
bool
UIFileChooser::handleUserInput(string_t input)
{
    // Split into file and directory part
    string_t dir  = Directory::getDirName(input);
    string_t file = Directory::getFileName(input);
    FileNamePattern npat(file);

    // Empty file name? Do nothing. Happens when users press Enter
    // on the input line.
    if (file.empty()) {
        return true;
    }

    // If this is just a pattern, re-list the directory, no matter which
    // (this allows changing the pattern while on the "roots" view)
    if (file == input && npat.hasWildcard()) {
        file_patterns.assign(1, npat);
        setDirectory(current_directory);
        return true;
    }

    // Handle relative file names
    if (!Directory::isAbsoluteName(dir)) {
        // Relative file names cannot be used in the "roots" view
        if (current_directory == 0) {
            messageBox(_("You cannot create files here. Please choose a place (drive, partition) first."),
                       _("File Selection"));
            return false;
        }
        dir = Directory::makeFileName(current_directory->getName(), dir);
    }

    // Remove ".." (getAbsoluteName is required to expand paths like "c:foo")
    dir = Directory::getAbsoluteName(dir);

    // Is this actually a directory?
    Ptr<Directory> ndir = Directory::openDir(dir);
    try {
        ndir->getEntries();
    }
    catch (FileProblemException& e) {
        messageBox(format(_("Unable to read directory %s: %s"), e.getFileName(), e.what()),
                   _("File Selection"));
        return false;
    }

    // Now, examine the file part
    if (npat.hasWildcard()) {
        // directory and wildcard given
        file_patterns.assign(1, npat);
        setDirectory(ndir);
    } else {
        // no wildcard. 
        Ptr<DirectoryEntry> e(ndir->getEntryByName(file));
        if (e->getFileType() == DirectoryEntry::FT_DIRECTORY) {
            // it's a directory
            setDirectory(e->openDir());
        } else {
            // assume it's a file
            result_file = e->getName();
            sig_file_selected.raise();
        }
    }
    return true;
}

/** Handle "Enter" action. When the cursor is on a directory, enters that.
    When the cursor is on a file, produce a result. */
void
UIFileChooser::handleEnter()
{
    if (getCurrent() < int(directories.size())) {
        setDirectory(directories[getCurrent()].directory);
    } else {
        result_file = files[getCurrent() - directories.size()]->getName();
        sig_file_selected.raise();
    }
}

/** Scan directory.
    \param directory Directory to read
    \param interactive true iff this is interactive operation */
void
UIFileChooser::scanDirectory(Ptr<Directory> directory, bool interactive)
{
    string_t error;

    /* Forget old content */
    directories.clear();
    files.clear();

    if (directory == 0) {
        /* List roots */
        Ptr<Enumeration<Ptr<Directory> > > items = Directory::getRoots();
        while (items->hasMoreElements()) {
            Ptr<Directory> d = items->getNextElement();
            directories.push_back(Dir(d->getTitle(), d, false));
        }
    } else {
        /* First, add "go up" entry */
        if (Ptr<Directory> p = directory->getParent()) {
            directories.push_back(Dir(_("[Up one level]"), p, true));
        } else {
            directories.push_back(Dir(_("[Other Places]"), 0, true));
        }

        /* Enumerate content */
        try {
            Ptr<Enumeration<Ptr<DirectoryEntry> > > items = directory->getEntries();
            while (items->hasMoreElements()) {
                Ptr<DirectoryEntry> e = items->getNextElement();
                switch (e->getFileType()) {
                 case DirectoryEntry::FT_FILE:
                    for (size_t i = 0, n = file_patterns.size(); i < n; ++i) {
                        if (file_patterns[i].match(e->getBaseName())) {
                            files.push_back(e);
                            break;
                        }
                    }
                    break;
                 case DirectoryEntry::FT_DIRECTORY:
                    directories.push_back(Dir(e->getBaseName(), e->openDir(), false));
                    break;
                 default:
                    break;
                }
            }
        }
        catch (FileProblemException& e) {
            /* We cannot refuse entering the directory at this point,
               but we can inform the user. However, we must do that
               after updating the widget, because at this point the
               data is inconsistent. */
            error = format(_("Unable to read directory %s: %s"), e.getFileName(), e.what());
        }

        /* Sort */
        std::sort(directories.begin()+1, directories.end(), isDirLT);
        std::sort(files.begin(), files.end(), isEntryLT);
    }

    /* Configure the UIScrollable */
    setCurrent(0);
    setTop(0);
    setSize(directories.size() + files.size());
    updateSize();
    drawWidget();
    sig_moved.raise(getId());

    /* Show error */
    if (interactive && error.size()) {
        messageBox(error, _("File Selection"));
    }
}

/** Show directory. */
void
UIFileChooser::showDirectory(Ptr<Directory> directory)
{
    // Skip if dir is null
    if (!directory) {
        return;
    }

    // Find both directories' levels
    int current_level = getLevel(current_directory);
    int dir_level     = getLevel(directory);

    // Adjust so that dir is exactly beneath current
    ++current_level;
    while (dir_level > current_level) {
        directory = directory->getParent();
        --dir_level;
    }

    // Match?
    if (dir_level == current_level) {
        for (std::vector<Dir>::size_type i = 0; i != directories.size(); ++i) {
            if (directories[i].directory != 0 && directories[i].directory->getName() == directory->getName()) {
                scrollTo(i, 1);
                break;
            }
        }
    }
}

void
UIFileChooser::updateSize()
{
    /* We want columns that are 15 normal ems wide */
    int want_column_width = font_ems[FONT_NORMAL] * 15;
    if (getExtent().w < want_column_width) {
        /* Less than one column */
        current_columns = 1;
        current_column_width = getExtent().w;
    } else {
        /* At least one. If there is a partial column remaining, split
           that among the current ones */
        current_columns = getExtent().w / want_column_width;
        current_column_width = (getExtent().w + current_columns-1) / current_columns;
    }

    /* Set page size for scroll bar */
    setPageSize(current_columns * (getExtent().h / font_heights[FONT_NORMAL]));
}

bool
UIFileChooser::isDirLT(const Dir& a, const Dir& b)
{
    return strCaseCmp(a.title, b.title) < 0;
}

bool
UIFileChooser::isEntryLT(const Ptr<DirectoryEntry>& a, const Ptr<DirectoryEntry>& b)
{
    return strCaseCmp(a->getBaseName(), b->getBaseName()) < 0;
}
